/* installpkg v 0.5a                                                             */
/* (c) 2001-2004 Michael Zolk                                              */
/*                                                                         */
/* Usable under BSD conditions                                             */
/*                                                                         */
/* 0.5 - fixed error when creating temp dir                                */
/* 0.4 - added support for DIR keyword                                     */
/*       now can set owner/group/mode for FILEs, CONFFILES, DIRs           */
/*       fixed getpkginfo: tabs or spaces can be used as separators for    */
/*         keywords & arguments                                            */
/*       added support for DEVICE keyword                                  */
/* 0.3 - changes to use Holger Veit's ux2util.dll                          */
/*       SYMLINK creation fixed                                            */
/* 0.2 - use UX2Util DLL                                                   */
/*       added symlink creation                                            */
/*       -s and -r options now work                                        */
/*       changed the getpkginfo procedure to avoid 'false positives'       */
/*         when searching for a keyword                                    */
/*       added new keyword CONFFILE                                        */
/* 0.1 - first release                                                     */

call RxFuncAdd 'SysLoadFuncs','RexxUtil','SysLoadFuncs'
call SysLoadFuncs

/* call RxFuncAdd 'UX2LoadFuncs','UX2Util','UX2LoadFuncs'
call UX2LoadFuncs */

call RxFuncAdd 'UX2CreateDir','UX2Util','UX2CreateDir'
call RxFuncAdd 'UX2UnixRoot','UX2Util','UX2UnixRoot'
call RxFuncAdd 'UX2FileExists','UX2Util','UX2FileExists'
call RxFuncAdd 'UX2MkSymlink','UX2Util','UX2MkSymlink'
call RxFuncAdd 'UX2Chown','UX2Util','UX2Chown'
call RxFuncAdd 'UX2Chgrp','UX2Util','UX2Chgrp'
call RxFuncAdd 'UX2Chmod','UX2Util','UX2Chmod'
call RxFuncAdd 'UX2Mknod','UX2Util','UX2Mknod'

/* find the TMP dir */
tmp=value('tmp',,'OS2ENVIRONMENT')
if tmp="" then
do
   say "You have to set the TMP env. variable"
   exit 1
end

/* parse options */
/* TODO: allow for multiple packages */

if arg()=0 then
do
  call usage
  exit
end

do i=1 to words(arg(1))
  argv.i=word(arg(1),i)
end

force="no"
verbose="no"
simul="no"
update="no"
root=""
argnum=1
do while argnum<=words(arg(1))-1
  if argv.argnum="-r" then
  do
    argnum=argnum+1
    root=argv.argnum
    argnum=argnum+1
  end
  else
  if argv.argnum="-f" then
  do
     force="yes"
     argnum=argnum+1
  end
  else
  if argv.argnum="-v" then
  do
      verbose="yes"
      argnum=argnum+1
  end
  else if argv.argnum="-s" then
  do
      simul="yes"
      argnum=argnum+1
  end
  else if argv.argnum="-v" then
  do
      update="yes"
      argnum=argnum+1
  end
  else   /* unknown option */
  do
    call usage
    exit
  end
end  /* while */

/* determine the target dir */
if root="" then unixroot=translate(UX2UnixRoot(),'\','/')
else unixroot=root

unixroot=translate(unixroot,'\','/')
unixroot=strip(unixroot,'T','\')

pkgstatedir=unixroot"\var\lib\unixos2\packages"
scrstatedir=unixroot"\var\lib\unixos2\scripts"

i=words(arg(1))
if left(argv.i,1) \= '-' then pkgfile=argv.i
else
do
  call usage
  exit 1
end

/* If someone left off the extension, try to figure that out: */
if UX2FileExists(pkgfile)=0 then
do
  if UX2FileExists(pkgfile".zip")=1 then pkgfile=pkgfile".zip" else
  if UX2FileExists(pkgfile".ux2")=1 then pkgfile=pkgfile".ux2"
end

/* are the base libs installed? */

if translate(pkgfile)\="OS2LIBS.ZIP" then
do
  call RxFuncAdd emx_revision, 'EMX', emx_revision
  signal on syntax name libemx_error
  dummy = emx_revision()
  signal off syntax
  call RxFuncDrop emx_revision
end

if UX2FileExists(pkgfile)=0 then
do
   say "File "pkgfile" not found"
   exit 1
end

/* Reject package if it does not end in '.zip'or .ux2 */
if translate(right(pkgfile,4)) \= '.ZIP' & translate(right(pkgfile,4)) \= '.UX2' then
do
  say "Cannot install "pkgfile": not a .zip or .ux2 file"
  exit 1
end

/* unpack the PKGINFO file */
"@unzip.exe -o "pkgfile" PKGINFO -d "tmp" >NUL:"

/* get the package name */
call getpkginfo "PKGNAME"
if pkginfo.0=1 then pkgname=pkginfo.1
else if pkginfo.0 \= 1 then
do
   say "error in PKGINFO file - multiple or missing PKGNAME(s)"
   if force="no" then exit 1 else pkgname="unknown_package"
end

/* output pkg information when in verbose mode */
if verbose="yes" then
do
   say "Installing "pkgname" from archive file "pkgfile
   call getpkginfo "DESC"
   if pkginfo.0>0 then
   do
       do i=1 to pkginfo.0
         say pkginfo.i
       end  /* do i=1 ... */
   end  /* if pkginfo.0... */
   call getpkginfo "CONTACT"
   if pkginfo.0>0 then
   do
     say ""
     do i=1 to pkginfo.0
       say pkginfo.i
     end
   end
end

/* abort installation if requirements are not met */
call getpkginfo "REQUIRES"
if pkginfo.0>0 then
do
   do i=1 to pkginfo.0
      reqpkg=word(pkginfo.i,1)
      if UX2FileExists(unixroot"\var\log\packages\"reqpkg) \= 1 then
      do
          /* required package not installed */
          say "required package "reqpkg" not installed"
          if force="no" then
          do
             say "installation aborted"
             exit 1
          end
          else say "installation forced - continuing"
      end /* if UX2FileExist... */
   end
end

dbfile=unixroot"\var\log\packages\"pkgname

/* dbfile exists -> package already installed */
if UX2FileExists(dbfile)=1 then
do
  say "Package "pkgname" is already installed."
  say "Use upgradepkg to upgrade an installed package."
  exit
end

/* execute PREINST script */
j=0
call getpkginfo "SCRIPT"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
     if word(pkginfo.i,1)="PREINST" then
     do
        j=j+1
        pkginfo.i=delword(pkginfo.i,1,1)
        "@unzip.exe -o "pkgfile" install\"word(pkginfo.i,1)" -d "unixroot"\ > NUL:"
        if verbose="yes" then say "executing PREINST script "pkginfo.i
        olddir=directory()
        call directory unixroot"\install"
        if simul="no" then
        do
          "@cmd /c "pkginfo.i
          /* copy the script to /var/log/scripts */
          "@copy "word(pkginfo.i,1)" "unixroot"\var\log\scripts\"pkgname".preinst"j" > nul:"
        end
        call directory olddir
     end /* if word() */
  end  /* do i=1... */

/* create directories */
call getpkginfo "DIR"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    dir=strip(translate(word(pkginfo.i,1),'\','/'),'L','\')
    dir=unixroot"\"dir
    if verbose="yes" then say "Creating directory "dir
    if simul="no" then call UX2CreateDir dir
    if words(pkginfo.i)>1 then
      if simul="no" then call setrights pkginfo.i,dir
  end

/* unzip the package: */
call getpkginfo "FILE"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    file=word(pkginfo.i,1)
    if verbose="yes" then say "unpacking "file
    /* depends on makepkg if the following is always necessary? */
    if left(file,1)="/" then file=strip(file,'L','/')
    /* TODO: handle locked files */
    if simul="no" then
    do
       "@unzip.exe -o "pkgfile" "file" -d "unixroot"\ > NUL:"
       if words(pkginfo.i)>1 then call setrights pkginfo.i,unixroot"\"file
    end
  end

/* extract configuration files */
call getpkginfo "CONFFILE"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    rename="no"
    file=word(pkginfo.i,1)
    if verbose="yes" then say "unpacking configuration file "file
    if left(file,1)="/" then file=strip(file,'L','/')
    if update="yes" then
    do
      if UX2FileExists(UX2Canonicalize("/"file)) then
      do
         say "Warning: configuration file "UX2Canonicalize("/"file)" already exists - unpacking "UX2BaseName(file)".new"
         "@ren "UX2Canonicalize("/"file)" "UX2Canonicalize("/"file".tmp")
         rename="yes"
      end
    end
    if simul="no" then
    do
      "@unzip.exe -o -n "pkgfile" "file" -d "unixroot"\ > NUL:"
      if rename="yes" then
      do
         "@ren "UX2Canonicalize("/"file)" "UX2Canonicalize("/"file".new")
         "@ren "UX2Canonicalize("/"file".tmp")" "UX2Canonicalize("/"file)
      end
      if words(pkginfo.i)>1 then call setrights pkginfo.i,unixroot"\"file
    end
  end

/* create device nodes */
call getpkginfo "DEVICE"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    devname=strip(translate(word(pkginfo.i,1),"\","/"),"L","\")
    type=word(pkginfo.i,2)
    major=word(pkginfo.i,3)
    minor=word(pkginfo.i,4)
    if verbose="yes" then say "Creating device node "devname
    if simul="no" then call UX2Mknod unixroot"\"devname,type,major,minor
  end

/* create symlinks */
call getpkginfo "SYMLINK"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    file=translate(word(pkginfo.i,1),'\','/')
    file=strip(file,'L','\')
    symlink=translate(word(pkginfo.i,2),'\','/')
    symlink=strip(symlink,'L','\')
    if verbose="yes" then say "creating symlink "symlink" -> "file
    if simul="no" then call UX2MkSymlink unixroot"\"symlink,unixroot"\"file
  end

/* execute POSTINST script */
j=0
call getpkginfo "SCRIPT"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
     if word(pkginfo.i,1)="POSTINST" then
     do
        j=j+1
        pkginfo.i=delword(pkginfo.i,1,1)
        "@unzip.exe -o "pkgfile" install\"word(pkginfo.i,1)" -d "unixroot"\ > NUL:"
        if verbose="yes" then say "executing POSTINST script "pkginfo.i
        olddir=directory()
        call directory unixroot"\install"
        if simul="no" then
        do
          "@cmd /c "pkginfo.i
          /* copy the script to /var/log/scripts */
          "@copy "word(pkginfo.i,1)" "unixroot"\var\log\scripts\"pkgname".postinst"j" > nul:"
        end
        call directory olddir
     end /* if word() */
  end  /* do i=1... */

/* write package database entry */
/* i.e. copy PKGINFO to /var/lib/unixos2/packages :-) */
if verbose="yes" then say "creating package database entry"
if simul="no" then "@copy "tmp"\PKGINFO "pkgstatedir"\"pkgname" > nul:"

/* store POSTDEL script for later use */
j=0
call getpkginfo "SCRIPT"
if pkginfo.0>0 then
  do i=1 to pkginfo.0
     if word(pkginfo.i,1)="POSTDEL" then
     do
        j=j+1
        pkginfo.i=delword(pkginfo.i,1,1)
        "@unzip -o "pkgfile" install\"word(pkginfo.i,1)" -d "tmp" >NUL:"
        olddir=directory()
        call directory tmp"\install"
        /* copy the script to /var/lib/unixos2/scripts */
        if simul="no" then "@copy "word(pkginfo.i,1)" "scrstatedir"\"pkgname".postdel"j" > nul:"
        call directory olddir
     end /* if word() */
  end  /* do i=1... */

if verbose="yes" then say "Installation successful."

endofinst:
exit

usage: procedure

say "Usage: installpkg [options] package_name"
say ""
say "Installpkg is used to install an UnixOS/2 package"
say ""
say "options:       -v        verbose"
say "               -f        force installation"
say "               -r dir    use 'dir' as root directory for installation"
say "               -s        simulate (do not install)"
return

/* extract all lines for a given keyword from the PKGINFO file */
/* writes the lines minus the keyword into the global stem variable pkginfo */
getpkginfo: procedure expose tmp pkginfo.

parse arg keyword

rc=SysFileSearch(keyword,tmp"\PKGINFO",'tmppkg.','C')
if rc \=0 then
  return 1
else
do
  pkginfo.0=0
  j=1
  if tmppkg.0>0 then
  do
    do i=1 to tmppkg.0
       tmppkg.i=translate(tmppkg.i," ",d2c(9))
       if word(tmppkg.i,1)=keyword then
       do
          pkginfo.j=delword(tmppkg.i,1,1)
          j=j+1
          pkginfo.0=pkginfo.0+1
       end  /* if word... */
    end   /* i=1 to tmppkg.0 */
  end /* if tmppkg.0... */
  return 0
end  /* else */

setrights: procedure

parse arg pkgline,file

owner=word(pkgline,2)
group=word(pkgline,3)
mode=word(pkgline,4)
call UX2Chown file,owner
call UX2Chgrp file,group
call UX2Chmod file,mode
return

libemx_error:

say "os2libs not found - installation aborted"
exit 1

