/* removepkg - UnixOS/2 package deinstaller                    */
/* (c) 2001 Michael Zolk                                       */
/*                                                             */
/* 0.4  bumped up version number to be in sync with installpkg */
/*      remove empty directories                               */
/*      remove device nodes                                    */
/*      fixed getpkginfo                                       */
/* 0.1  initial release                                        */
/*                                                             */
/* usable under BSD conditions                                 */

call RxFuncAdd 'SysLoadFuncs','RexxUtil','SysLoadFuncs'
call SysLoadFuncs

call RxFuncAdd 'UX2LoadFuncs','UX2Util','UX2LoadFuncs'
call UX2LoadFuncs

/* find the TMP dir */
tmp=value('tmp',,'OS2ENVIRONMENT')
if tmp="" then
do
   say "You have to set the TMP env. variable"
   exit 1
end

/* parse options */
/* TODO: allow for multiple packages */

if arg()=0 then
do
  call usage
  exit
end

do i=1 to words(arg(1))
  argv.i=word(arg(1),i)
end

force="no"
verbose="no"
simul="no"
update="no"
root=""
presdir=""
argnum=1
do while argnum<=words(arg(1))-1
  if argv.argnum="-r" then
  do
    argnum=argnum+1
    root=argv.argnum
    argnum=argnum+1
  end
  else
  if argv.argnum="-s" then
  do
     argnum=argnum+1
     presdir=argv.argnum
     argnum=argnum+1
  end
  else
  if argv.argnum="-f" then
  do
     force="yes"
     argnum=argnum+1
  end
  else
  if argv.argnum="-v" then
  do
      verbose="yes"
      argnum=argnum+1
  end
  else if argv.argnum="-s" then
  do
      simul="yes"
      argnum=argnum+1
  end
  else if argv.argnum="-u" then
  do
      update="yes"
      argnum=argnum+1
  end
  else   /* unknown option */
  do
    call usage
    exit
  end
end  /* while */

i=words(arg(1))
if left(argv.i,1) \= '-' then pkgfile=argv.i
else
do
  call usage
  exit 1
end
pkgfile=UX2BaseName(pkgfile)

/* determine the target dir */
if root="" then unixroot=value('unixroot',,'OS2ENVIRONMENT')
else unixroot=root
unixroot=translate(unixroot,'\','/')
unixroot=strip(unixroot,'T','\')

/* check if the package is installed */
if UX2FileExists(unixroot"\var\log\packages\"pkgfile)=0 then
do
  say "No such package: "unixroot"\var\log\packages\"pkgfile". Can't remove."
  exit 1
end

if verbose="yes" then say "Removing package "unixroot"\var\log\packages\"pkgfile

/* check if package to be removed is required by any other package */
if update="no" then
do
   call SysFileTree unixroot"\var\log\packages\*","file.","FO"
   if file.0>0 then   /* should always be true */
     do i=1 to file.0
       call getpkginfo "REQUIRES",file.i
       if pkginfo.0>0 then
         do i=1 to pkginfo.0
           if translate(word(pkginfo.i,1))=translate(pkgfile) then    /* package is required */
           do
             say "Package "pkgfile" is required by "UX2BaseName(file.i)
             if force="no" then exit 1
           end

         end /* i=1 to pkginfo.0 */

     end /* i=1 to file.0 */
end

/* remove the symlinks */
call getpkginfo "SYMLINK",unixroot"\var\log\packages\"pkgfile
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    symlink=translate(word(pkginfo.i,2),'\','/')
    symlink=strip(symlink,'L','\')
    if verbose="yes" then say "Removing symlink "symlink" pointing to "word(pkginfo.i,1)
    if simul="no" then call UX2DelSymLink unixroot"\"symlink
    /* TODO: what should happen to symlinks when using -p option? */
  end

/* remove device nodes */
call getpkginfo "DEVICE",unixroot"\var\log\packages\"pkgfile
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    devname=strip(translate(word(pkginfo.i,1),"\","/"),"L","\")
    if verbose="yes" then say "Removing device node "devname
    if simul="no" then call SysFileDelete(devname)
  end

/* remove the files */
call getpkginfo "FILE",unixroot"\var\log\packages\"pkgfile
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    filename=word(pkginfo.i,1)
    filename=translate(filename,'\','/')
    filename=strip(filename,'L','\')
    if presdir="" then
      action="Removing "
    else
      action="Preserving "
    if verbose="yes" then say action||filename
    if simul="no" then
    do
      if presdir="" then
        call SysFileDelete unixroot"\"filename
      else
        "@cmd.exe /c move "unixroot"\"filename" "presdir
    end
  end

/* the same for CONFFILEs */
if update="no" then
do
   call getpkginfo "CONFFILE",unixroot"\var\log\packages\"pkgfile
   if pkginfo.0>0 then
     do i=1 to pkginfo.0
       filename=word(pkginfo.i,1)
       filename=translate(filename,'\','/')
       filename=strip(filename,'L','\')
       if presdir="" then
         action="Removing "
       else
         action="Preserving "
       if verbose="yes" then say action"configuration file "filename
       if simul="no" then
       do
         if presdir="" then
           call SysFileDelete unixroot"\"filename
         else
           "@cmd.exe /c move "unixroot"\"filename" "presdir" > NUL:"
       end
     end
end

/* remove directories if they're empty */
call getpkginfo "DIR",unixroot"\var\log\packages\"pkgfile
if getpkginfo.0>0 then
  do i=1 to pkginfo.0
    dir=word(pkginfo.i,1)
    if root="" then
      dir=UX2Canonicalize(dir)
    else
      dir=unixroot"\"dir
    call SysFileTree translate(dir"\*",'\','/'),'file.'
    if file.0=0 then
    do
        if verbose="yes" then say "Removing empty directory "dir
        if simul="no" then "@cmd.exe /c rmdir "translate(dir,'\','/')
    end
    else say "Directory "dir" not removed - not empty"
  end

/* run the postdel script */
call getpkginfo "SCRIPT",unixroot"\var\log\packages\"pkgfile
j=0
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    if word(pkginfo.i,1)="POSTDEL" then
    do
      j=j+1
      if verbose="yes" then say "executing postdel script "delword(pkginfo.i,1,1)
      if simul="no" then
      do
        "@cmd.exe /c copy "unixroot"\var\log\scripts\"pkgfile".postdel"j" "tmp"\"word(pkginfo.i,2)" >NUL:"
        curdir=directory()
        call directory tmp
        "@cmd.exe /c "delword(pkginfo.i,1,1)
        call directory curdir
      end
    end
  end

/* copy the scripts to presdir or kill 'em all */
postdel=0
preinst=0
postinst=0
if pkginfo.0>0 then
  do i=1 to pkginfo.0
    select
      when word(pkginfo.i,1)="PREINST" then
      do
        preinst=preinst+1
        oldname=unixroot"\var\log\scripts\"pkgfile".preinst"preinst
      end
      when word(pkginfo.i,1)="POSTINST" then
      do
        postinst=postinst+1
        oldname=unixroot"\var\log\scripts\"pkgfile".postinst"postinst
      end
      when word(pkginfo.i,1)="POSTDEL" then
      do
        postdel=postdel+1
        oldname=unixroot"\var\log\scripts\"pkgfile".postdel"postdel
      end
    end /* select */
    if presdir="" then
      action="Removing "
    else
      action="Preserving "
    if verbose="yes" then say action||oldname
    if simul="no" then
      if presdir="" then
        call SysFileDelete oldname
      else
        "@cmd.exe /c copy "oldname" "presdir"\"word(pkginfo.i,2)" > NUL:"
  end /* i=1 to pkginfo.0 */

/* remove the pkg db entry */
if verbose="yes" then say "Removing package database entry for "pkgfile
if simul="no" then
  if presdir="" then
    call SysFileDelete unixroot"\var\log\packages\"pkgfile
  else
    "@cmd.exe /c copy "unixroot"\var\log\packages\"pkgfile" "presdir"\PKGINFO > NUL:"

exit
usage: procedure

say "Usage: removepkg [options] packagename"
say ""
say "removepkg is used to remove an UnixOS/2 package"
say ""
say "options:   -v      verbose"
say "           -f      force deinstallation"
say "           -r dir  use 'dir' as root dir"
say "           -s      simulate (do not remove anything)"
say "           -p dir  preserve removed files in 'dir'"
return

/* extract all lines for a given keyword from the PKGINFO file */
/* writes the lines minus the keyword into the global stem variable pkginfo */
getpkginfo: procedure expose tmp pkginfo. unixroot

parse arg keyword,pkginfofile

rc=SysFileSearch(keyword,pkginfofile,'tmppkg.','C')
if rc \=0 then
  return 1
else
do
  pkginfo.0=0
  j=1
  if tmppkg.0>0 then
  do
    do i=1 to tmppkg.0
       tmppkg.i=translate(tmppkg.i," ",d2c(9))
       if word(tmppkg.i,1)=keyword then
       do
          pkginfo.j=delword(tmppkg.i,1,1)
          j=j+1
          pkginfo.0=pkginfo.0+1
       end  /* if word... */
    end   /* i=1 to tmppkg.0 */
  end /* if tmppkg.0... */
  return 0
end  /* else */

