/* makepkg  Program to create a UnixOS/2 package file. */
/* Copyright (c) 2003  Andreas Buening */
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* 
**********************************************************************
***** constants ******************************************************
**********************************************************************
*/
/* program specific constants */
const.PROGRAM = 'makepkg'
const.VERSION = '0.1.1'
const.DESCRIPTION = 'Program to create a UnixOS/2 package file'
Pk = const.PROGRAM const.VERSION || '  ' || const.DESCRIPTION
CopyRight = 'Copyright (c) 2003  A. Buening'
Revision = 'Date of last change: 30 Aug 2003'

const.HelpInfo = "Please enter '--help' to get usage information"
const.Usage = const.PROGRAM '[ OPTIONS ] | --help'
const.ValidName = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-' /* valid characters */
const.ValidVersion = '0123456789.'

/* generic constants */
true = 1
false = 0
CL = D2C(13)
LF = D2C(10)
Tab = D2C(9)



/* ***** List of all global constants ******************************** */
CONSTS = 'const. Pk CopyRight Revision true false CL Tab'

/*
**********************************************************************
***** Global variables ***********************************************
**********************************************************************
*/

/* program options */
/* general program options */
option.batch = 0      /* batch mode: succeed or fail but never ask the user */
option.debug = 0      /* extra debug info */
option.force = 0      /* --force */
option.list = 0       /* list/demo mode */
option.unixroot = ''  /* the unixroot drive */
option.quiet = 0      /* quiet mode */
option.verbose = 0    /* verbose mode */
option.wait = 0       /* wait for key pressed before terminating the program */
option.tmpdir = ''    /* the tmp directory */


/* global variables */
global.cwd = ''               /* the current working directory */
file.count = 0
dir.count = 0
package.count = 0
package.required.count = 0
package.recommended.count = 0
package.uninstall.count = 0   /* UNINSTALL package */
package.dir.count = 0         /* install dir */
package.file.count = 0        /* install file */
/* package.file.i.{name, action, type, md5sum, shared} */

package.install.count = 0     /* FILE INSTALL file */
package.info.count = 0        /* FILE INFO file */
package.man.count = 0         /* FILE MAN file */
package.noinstall.count = 0   /* FILE UNINSTALL file */
package.unknown.count = 0     /* ??? file */
package.ignore.count = 0      /* #FILE INSTALL file */
package.locale.count = 0      /* Locale message file */


/* ***** List of all global variables ****************** */
VARS = 'option. global. dir. file. package.'


/* ***** Exception handling ****************************************** */
/* It is possible to catch all exceptions that are activated by "CALL ON".
   E.g.:Define Catch.Error to 1 if you want to catch the "Error" exception
   just before the critical code section. Define it to 2 if you also want to
   prevent the exception handler from writing any message to stderr.
   In this case the error code is returned in "rc". */
SIGNAL ON SYNTAX   NAME Fehler.Syntax       /* Exception for "syntax error" */
SIGNAL ON NOVALUE  NAME Exception.NoValue   /* Exception for "undefined variable" */
CALL   ON ERROR    NAME Exception.Error     /* External program failed. */
CALL   ON FAILURE  NAME Exception.Failure   /* Unknown external command. */
CALL   ON HALT     NAME Exception.Halt      /* Ctrl+C */
CALL   ON NOTREADY NAME Exception.NotReady  /* I/O error */



main: /* ***** The main program ***** */
  /* char *argv */
  ret = SetLocal() /* Save the current environment. */

  /* Parse the command line and save the result in argc, argv. */
  parse arg commandline
  call ParseCommandLine commandline
  drop commandline

  /* Load REXXUTIL.DLL */
  call Load_REXXUTIL

  /* Get current directory */
  global.cwd = Directory()
  if length(global.cwd) <= 3 then do
    /* createpkg can not be run in the root directory */
    call Terminate 1, PROGRAM || ' can not be run in the root directory'
  end
  
  /* Ask for basic package info and write the result into package. */
  call ux2_ReadPackageInfo
  
  /* Read package files */
  call ux2_ReadPackageFiles

  /* Process package files */
  call ux2_ProcessPackageFiles

  /* Write the info into PKGINFO */
  call ux2_WritePackageInfo

  /* wait for ENTER */
  if option.wait then do
    say 'Press ENTER to continue ...'
    Ch = linein(,,)
  end

  ret = EndLocal() /* Restore the environment. */
exit /* Ende von main */


/*
*********************************************************************
***** Functions of makepkg ******************************************
*********************************************************************
*/

ux2_ReadPackageInfo: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Asks for the package info */
  say "Preparing package in directory '" || Directory() || "'"
  call charout 'stdout:', 'Enter the package name: '
  package.name = linein()

  if length(package.name) = 0 | CheckValidName(package.name, const.ValidName) = false then do
    call lineout 'stderr:', 'Package name contains invalid characters.'
    call Terminate 1, "Allowed are: A-Z,a-z,0-9,'_' and '-'."
  end

  call charout 'stdout:', 'Enter the package version: '
  package.version = linein()
  i = CheckValidVersion(package.version)
  if i < 0 then do
    call lineout 'stderr:', 'Package version is invalid. Version  is n.n.n[.n]x.'
    call Terminate 1, "n is a digit (0-9), x may be any string (a-z,A-Z,0-9,'_','-')"
  end

  package.date = date('S') 
  say "package date is '" || package.date || "'"

  call charout 'stdout:', 'Enter the package type (CORE/USR/LOCAL/OPT): '
  package.type = linein()
  parse upper var package.type package.type
  if package.type <> 'CORE' & package.type <> 'USR' & package.type <> 'LOCAL' &,
    package.type <> 'OPT' then call Terminate 1, "Unknown package type " || package.type

  call charout 'stdout:', 'Enter the contact info: '
  package.contact = linein()

  call charout 'stdout:', 'Required packages (space separated list): '
  string = linein()
  do while length(string) <> 0
    parse var string pkg string
    if CheckValidName(pkg, const.ValidName) then
      call ux2_AddEntry 'package.required', pkg
    else
      call Terminate 1, "Package name '" || pkg || "' is invalid."
  end /* end of do until length(string) = 0 */

  call charout 'stdout:', 'Recommended packages: '
  string = linein()
  do while length(string) <> 0
    parse var string pkg string
    if CheckValidName(pkg, const.ValidName) then
      call ux2_AddEntry 'package.recommended', pkg
    else
      call Terminate 1, "Package name '" || pkg || "' is invalid."
  end /* end of do until length(string) = 0 */

  call charout 'stdout:', 'Packages to be uninstalled when this package is installed: '
  string = linein()
  do while length(string) <> 0
    parse var string pkg string
    if CheckValidName(pkg, const.ValidName) then
      call ux2_AddEntry 'package.uninstall', pkg
    else
      call Terminate 1, "Package name '" || pkg || "' is invalid."
  end /* end of do until length(string) = 0 */
return


ux2_ReadPackageFiles: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Reads all package files and stores them in dir. and file. */
  /* Read all subdirectories. */
  rc = SysFileTree('*', 'dir', 'DSO')
  if rc <> 0 then call Terminate 1, 'Can not read directory information'
  dir.count = dir.0

  /* Remove the leading current directory from all directory names */
  len = length(global.cwd)
  do i = 1 to dir.count
    dir.i = substr(dir.i, len + 2)
  end

  /* print some info */
  if option.quiet = 0 then do
    if option.verbose then do
      do i = 1 to dir.count
        say 'Directory found:' dir.i
      end
    end
    say 'Number of subdirectories:' dir.count
  end
  if dir.count = 0 then
    call lineout 'stderr:', 'WARNING: No subdirectories found'

  /* Read all files in subdirectories */
  rc = SysFileTree('*', 'file', 'FSO')
  if rc <> 0 then call Terminate 1, 'Can not read file information'
  file.count = file.0

  /* Remove the leading current directory from all file names */
  len = length(global.cwd)
  do i = 1 to file.count
    file.i = substr(file.i, len + 2)
  end

  /* print some info */
  if option.quiet = 0 then do
    if option.verbose then do
      do i = 1 to file.count
        say 'File found:' file.i
      end
    end
    say 'Number of files:' file.count
  end
  if file.count = 0 then
    call lineout 'stderr:', 'WARNING: No files found'
return


ux2_ProcessPackageFiles: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Find out which file is what. */

  /* Add all directories to the package */
  package.dir.count = dir.count
  do i = 1 to dir.count
    package.dir.i = '/' || ChangeChar(dir.i, '\', '/')
  end

  /* Now add the files. */
  do i = 1 to file.count
    /* divide the file name into "dirname" and "basename". */
    file.i = '/' || ChangeChar(file.i, '\', '/')
    dirname = Dirname(file.i)
    basename = Basename(file.i)
    if option.debug = 1 then
      call lineout 'stderr:', "processing file '" || file.i || "' ('" || dirname || "', '" || basename || "')"

    if length(dirname) = 0 then do
      /* Case 1: files in the "root" directory of the package are special files. */
      parse upper var basename UpperBaseName

      select
        when UpperBaseName = 'COPYING' | UpperBaseName = 'README',
          | UpperBaseName = 'INSTALL' | UpperBaseName = 'PKGINFO' then
          call ux2_AddPackageFile 'package', basename, 'NOINSTALL'

        otherwise call ux2_AddPackageFile 'package', '/' || basename, 'NOINSTALL'
      end /* end of select */
    end /* end of if length(dirname) = 0 */

    else do /* length(dirname) <> 0 */
      /* Case 2: Analyse files in subdirectories. */
      type = ux2_CheckFileType(dirname, basename)
      
      select
        when type = 'ignore' then
          call ux2_AddPackageFile 'package', file.i, 'NOINSTALL'

        when type = 'info' then
          call ux2_AddPackageFile 'package', file.i, 'INSTALL', 'INFO'

        when type = 'man' then
          call ux2_AddPackageFile 'package', file.i, 'INSTALL', 'MAN'

        otherwise do
          call ux2_AddPackageFile 'package', file.i
        end

      end /* end of select */
    end /* end of else do */

  end /* end of do i = 1 to file.count */

return


ux2_AddPackageFile: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Adds the file 'filename' to the compound variable 'package'. */
  parse arg varname, filename, action, type, md5sum, shared

  if varname = '' | filename = '' then
    call Terminate 1, 'ux2_AddFile: syntax error'

  /* Replace '\' by '/' */
  filename = ChangeChar(filename, '\', '/')

  /* to upper case */
  parse upper var action action
  parse upper var type type
  parse upper var shared shared

  /* Set the default values */
  if action = '' then action = 'INSTALL' /* default action: install file */
  if type = '' then type = 'DEFAULT'
  if md5sum = '' then md5sum = 'NONE'
  if shared = '' then shared = 'NOSHARED'

  /* Search the file; if it does not yet exist add a new entry. */
  i = ux2_SearchEntry(varname || '.file', filename)
  i = 0
  if i = 0 then do
    i = ux2_AddEntry(varname || '.file', filename)
    /* Initialize with default values. */
    command = varname || '.file.i.install = 0'
    interpret command
    command = varname || '.file.i.uninstall = false'
    interpret command
  end

  /* Add the action */
  select
    when action = 'INSTALL' then command = 'INSTALL = 1'
    when action = 'NOINSTALL' then command = 'INSTALL = 2'
    when action = 'UNINSTALL' then command = 'UNINSTALL = true'
  end
  command = varname || '.file.i.' || command
  interpret command

  if action = 'INSTALL' then do
    /* These infos are only valid for 'INSTALL' */
    command = varname || '.file.i.type = type'
    interpret command
    command = varname || '.file.i.md5sum = md5sum'
    interpret command
    command = varname || '.file.i.shared = shared'
    interpret command
  end

  return true


ux2_WritePackageInfo: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Write package information to PKGINFO */
  Ch = AskUser('Do you want to write the package information into PKGINFO (y/n)? ', 'YN')
  if Ch = 'Y' then do
    /* Check whether the output file already exists. */
    rc = Stream('PKGINFO', 'C', 'QUERY EXISTS')
    if rc <> '' then do
      rc = ux2_Remove('PKGINFO')
      if rc <> 0 then call Terminate 1, "Can not remove file 'PKGINFO'"
    end

    rc = ux2_WritePKGINFO('PKGINFO')
    if rc <> 0 then
      call Terminate 1, 'Can not write PKGINFO file'
    else do
      if option.debug = 0 then
        call lineout 'stderr:', 'PKGINFO successfully written'
    end

  end /* end of if Ch = 'Y' */
return


ux2_WritePKGINFO: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Writes the data into the file specified by argument 1. */
  parse arg handle

  rc = Stream(handle, 'C', 'QUERY EXISTS')
  if rc <> '' then do
    if option.debug = 1 then
      call lineout 'stderr:', "File '" || handle || "' already exists"
    return 1
  end

  /* In List mode just return 0. */
  if option.list = 1 then do
    if option.quiet = 0 then
      call lineout 'stdout:', 'List mode: nothing to do'
    return 0
  end

  rc = Stream(handle, 'C', 'OPEN WRITE')
  if rc = 'READY:' then do
    if option.verbose = 1 then
      call lineout 'stdout:', "Writing file '" || handle || "' ..."

    call lineout handle, '# This file was generated by ' || const.PROGRAM || ' ' || const.VERSION
    call lineout handle, 'PKGNAME ' || package.name
    call lineout handle, 'VERSION ' || package.version
    call lineout handle, 'DATE ' || package.date
        if package.contact <> '' then
      call lineout handle, 'CONTACT ' || package.contact
    
    do i = 1 to package.required.count
      call lineout handle, 'REQUIRED ' || package.required.i
    end

    do i = 1 to package.uninstall.count
      call lineout handle, 'UNINSTALL ' || package.uninstall.i
    end

    do i = 1 to package.dir.count
      call lineout handle, 'DIR ' || package.dir.i || ' bin bin 444'
    end

    do i = 1 to package.file.count
      if package.file.i.uninstall then
        call lineout handle, 'FILE UNINSTALL ' || package.file.i
      if package.file.i.install = 1 then
        call lineout handle, 'FILE ' || package.file.i || ' bin bin 444'
      else if package.file.i.install = 2 then
        call lineout handle, '# FILE NOINSTALL ' || package.file.i
    end
    
    rc = Stream('PKGINFO', 'C', 'CLOSE')
  end

  if rc = 'READY:' then rc = 0
  else rc = 1 /* error */
return rc




/*
*********************************************************************
***** Standard installpkg functions *********************************
*********************************************************************
*/

AskUser: PROCEDURE
/* Asks the user for confirmation. 1st argument is printed, 2nd argument contains
   all characters that are valid as answer. */
  parse arg message, answer
  parse upper var answer answer /* to uppercase */

  do forever
    call charout 'stdout:', message
    /* char Ch; */
    /* Read one line (uppercase) and store the data in Ch. */
    parse upper value linein(,,) with Ch

    /* Ch must contain exactly one character. */
    if length(Ch) <> 1 then iterate

    i = Pos(Ch, answer)
    if (i <> 0) then leave /* Found */
  end /* do forever */

return Ch


Basename: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Returns the basename of a given path. */
  parse arg filespec

  filespec = ChangeChar(filespec, '/', '\')
  result = filespec('N', filespec)
return result


ChangeChar: PROCEDURE
/* Replaces single characters */
  parse arg string, old, new
  
  do until i = 0
    i = pos(old, string)
    if i <> 0 then string = overlay(new, string, i)
  end
  
return string /* End of ChangeChar */


CheckDrive: PROCEDURE EXPOSE (CONSTS)
/* Checks whether the argument is a drive letter. */
  parse arg drive

  /* Check whether UNIXROOT is valid */
  valid = true
  if length(drive) = 2 then do
    /* error if the first letter is not a-z or A-Z. */
    if datatype(substr(drive,1,1), 'M') = 0 then valid = false
    /* error if the second letter is not ':'. */
    if substr(drive,2,1) <> ':' then valid = false
  end
  else valid = false

return valid /* End of is_dir */


CheckValidName: PROCEDURE EXPOSE (CONSTS)
/* Checks whether the first argument consists totally of characters contained in
   the second argument */
  parse arg string, list
  parse upper var string string /* upper case */

  do i = 1 to length(string)
    ch = substr(string, i, 1) /* get character i */
    if pos(ch, list) = 0 then return false /* ch is not in list */
  end /* end of do i = 1 */

return true


CheckValidVersion: PROCEDURE EXPOSE (CONSTS)
/* Checks whether the argument is a valid version number.
   Result: valid version (0), invalid version (-1). A positive value defines
   the index of the first invalid character and is >= 2. This can e.g. happen
   for "1.2.3a". In this case the "a" is an invalid character. */
  parse arg ver

  /* Split the version number into pieces */
  parse var ver Major '.' Minor '.' Revision '.' tail '.' dummy

  /* At most 3 periods are allowed so dummy must be empty */
  if length(dummy) <> 0 then return -1

  /* If there are only 2 periods then tail is empty. Assume Revision = 0. */
  if length(tail) = 0 then do
    tail = Revision
    Revision = '0'
  end

  /* If there is only 1 period the tail is still empty. Assume Minor = 0. */
  if length(tail) = 0 then do
    tail = Minor
    Minor = '0'
  end

  /* If ver is still empty then there is no period. */
  if length(tail) = 0 then return -1

  if CheckValidName(Major, '0123456789') = false then return -1
  if CheckValidName(Minor, '0123456789') = false then return -1
  if CheckValidName(Revision, '0123456789') = false then return -1

  /* Now check the content after the last period. The first character must
     be a digit. */
  invalid = 0
  do i = 1 to length(tail)
    ch = substr(tail, i, 1) /* character i of ver */
    if pos(ch, '0123456789') = 0 then do
      invalid = i /* character i is the first invalid one */
      leave
    end
  end /* end of do i = 1 to length(ver) */

  if invalid = 1 then return -1 /* first character must be valid */
  else if invalid > 1 then invalid = pos(ch, ver)
return invalid


Dirname: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Returns the dirname of a given path including the drive letter. */
  parse arg filespec

  filespec = ChangeChar(filespec, '/', '\')
  drive = filespec('D', filespec)
  dir = filespec('P', filespec)
  len = length(dir)
  if len = 0 then do
    len = 2
    dir = '.'
  end

  /* note: dir contains a trailing '\', so we have to remove it. */
  result = drive || substr(dir, 1, len - 1)
return result


ux2_AddEntry: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Adds the string in argument 2 to the list in argument 1, i.e.:
   i = VARIABLE.count + 1
   VARIABLE.count = i
   VARIABLE.i = STRING
   Returns the index of the new entry. */
  parse arg variable, string

  command = variable || '.count = ' || variable || '.count + 1'
  if option.debug = 1 then
    call lineout 'stderr:', "Interpreting command '" || command || "'"
  interpret command
  command = 'i = ' || variable || '.count'
  if option.debug = 1 then
    call lineout 'stderr:', "Interpreting command '" || command || "'"
  interpret command
  command = variable || '.' || i || ' = string'
  if option.debug = 1 then
    call lineout 'stderr:', "Interpreting command '" || command || "'"
  interpret command

return i


ux2_CheckFileType: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Finds out file type. */
  parse arg dirname, basename

  dir = substr(reverse(dirname), 2)  /* remove the last character from dirname */
  base = substr(basename, length(basename) - 1, 1) /* get the prelast character from basename */

  select
    /* Check whether dirname = '.../info' and basename = dir */
    when basename = 'dir' & substr(reverse(dirname), 1, 5) = "ofni/" then do
      result = 'ignore'
      if option.debug = 1 then
        call lineout 'stderr:', "file '" || dirname || '/' || basename || "' has to be ignored."
    end

    /* Check whether dirname = '.../info' and basename = '*.info' */
    when substr(reverse(dirname), 1, 5) = "ofni/" & substr(reverse(basename), 1, 5) = 'ofni.' then do
    /* Note: Extra files like foo.info-1, foo.info-2 must not be considered as .info
       files here because only foo.info needs a call of install-info.exe. */
      result = 'info'
      if option.debug = 1 then
        call lineout 'stderr:', "file '" || dirname || '/' || basename || "' is an info file."
    end

    /* Check whether dir = '.../man/man' and base = '*.?' */
    when base = '.' & substr(dir, 1, 8) = 'nam/nam/' then do
      result = 'man'
      if option.debug = 1 then
        call lineout 'stderr:', "file '" || dirname || '/' || basename || "' is a man page."
    end
    
    otherwise do
      result = 'other'
      if option.debug = 1 then
        call lineout 'stderr:', "file '" || dirname || '/' || basename || "' is a normal file."
    end
  end /* select */
  
  return result


GetTmpdir: PROCEDURE EXPOSE (CONSTS)
/* Delivers the current unixroot or an empty string if an error occurrs.
   Uses the OS/2 environment if TMPDIR is not given as argument.
   Checks whether the string is valid.
*/
  parse arg tmpdir
  
  /* Get TMPDIR from environment if not given as argument. */
  if tmpdir = '' then tmpdir = value('TMPDIR',,'OS2ENVIRONMENT')
  
  /* Check whether TMPDIR is valid */
  valid = true
  if length(tmpdir) > 3 then do
    /* TMPDIR must contain slashes not backslashes */
    if pos('\', tmpdir) <> 0 then valid = false
    /* first 2 bytes must be a drive letter */
    if CheckDrive(substr(tmpdir, 1, 2)) = false then valid = false
    /* third one must be a slash */
    if substr(tmpdir, 3, 1) <> '/' then valid = false
  end
  else valid = false
  
  if valid = false then tmpdir = ''
return tmpdir /* End of CheckTmpdir */


CheckUnixroot: PROCEDURE EXPOSE (CONSTS)
/* Delivers the current unixroot or an empty string if an error occurrs.
   Uses the OS/2 environment if UNIXROOT is not given as 1st argument.
   Checks whether the string is valid.
*/
  parse arg unixroot

  /* Get UNIXROOT from environment if not given as argument. */
  if unixroot = '' then unixroot = value('UNIXROOT',,'OS2ENVIRONMENT')

  /* Check whether UNIXROOT is valid */
  if CheckDrive(unixroot) = false then unixroot = '' /* delete unixroot */
return unixroot /* End of CheckUnixroot */


Load_REXXUTIL: PROCEDURE EXPOSE (VARS)
/* Loads the REXXUTIL.DLL. */
  rc = RxFuncAdd("SysLoadFuncs", "REXXUTIL", "SysLoadFuncs")
  if rc <> 0 & option.debug <> 0 then
    call lineout 'stderr:', 'WARNING: RxFuncAdd returned ' || rc

  /* Check whether SysLoadFuncs is available. The return code of
     RxFuncAdd is not sufficient. */
  rc = RxFuncQuery("SysLoadFuncs")
  if rc <> 0 then
    call Terminate rc, 'Could not load REXXUTIL.DLL'
  else do
    if option.verbose <> 0 then
      say 'REXXUTIL.DLL loaded'
  end
return


ux2_remove: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Removes a file. */
  parse arg filename

  /* Don't do anything in list mode */
  if option.list = true then rc = 0
  else rc = SysFileDelete(filename)

  if rc <> 0 then
    call lineout 'stderr:', "can not delete file '" || filename || "'"
  else if option.verbose = true then
    call lineout 'stdout:', "removing file '" || filename || "'"
return rc


ux2_SearchEntry: PROCEDURE EXPOSE (CONSTS) (VARS)
/* Searches for STRING in the compound variable VARNAME and returns its index
   or 0 if not found */
  parse arg varname, string

  command = 'n = ' || varname || '.count'
  interpret command

  result = 0
  i = 0
  do while i < n & result = 0
    i = i + 1
    command = 'temp = ' || varname || '.i'
    interpret command
    if temp = string then result = i /* entry found */
  end

  if option.debug then
    call lineout 'stderr:', "Searching for entry '" || string || "': " || result
return result




/*
*********************************************************************
***** Standard functions of the program *****************************
*********************************************************************
*/

ParseCommandLine: PROCEDURE EXPOSE argc argv. (CONSTS) (VARS)
/* Parses the command line. */
  /* char *commandline;
     int  argc;
     char *argv;
     int  i;
  */
  parse arg commandline
  parse source . . argv.0  /* file specification of the program itself */
  i = 0
  do while LENGTH(commandline) <> 0
    i = i + 1
    parse var commandline argv.i commandline  /* Save parameter i in argv.i */
  end
  argc = i + 1  /* like in C: (argvargc == NULL) */

  /* Call DisplayHelp or DisplayVersion if necessary */
  
  if argc = 2 then do
    if argv.1 = '/H' | argv.1 = '/h' | argv.1 = '/?' | argv.1 = '-H' | argv.1 = '-h',
               | argv.1 = '-?' | argv.1 = '--help'
      then call DisplayHelp
    if argv.1 = '/V' | argv.1 = '/v' | argv.1 = '-V' | argv.1 = '-v' | argv.1 = '--version'
      then call DisplayVersion
  end


  /* Parse the command line options */
  i = 1
  do while i < argc
    select
      
      /* general program options */
      when argv.i = '--batch' | argv.i = '-b' then option.batch = 1

      when argv.i = '--debug' | argv.i = '-d' then option.debug = 1

      when argv.i = '--force' | argv.i = '-f' then option.force = 1

      when argv.i = '--list' | argv.i = '-l' then option.list = 1

      when argv.i = '--quiet' | argv.i = '-q' then option.quiet = 1

      when argv.i = '--root' | argv.i = '-r' then do
        i = i + 1
        if i < argc then do
          option.unixroot = argv.i
        end
        else call Terminate 1, 'Syntax error: unixroot is missing'
      end

      when argv.i = '--verbose' | argv.i = '-v' then option.verbose = 1

      when argv.i = '--wait' | argv.i = '-w' then option.wait = 1


      /* special program options */

      otherwise call Terminate 1, 'unknown option ' || argv.i

    end /* select */

    i = i + 1
  end /* do while */

  
if option.quiet = false then do
    say Pk
    if option.list then
      say 'Running in demo mode: No changes are performed.'
  end

  /* check whether the options are consistent. */
  if option.batch = 1 & option.wait = 1 then
    call Terminate 1, 'Syntax error: --batch and --wait can not be combined.'

  if option.quiet = 1 & option.verbose = 1 then
    call Terminate 1, 'Syntax error: --quiet and --verbose can not be combined.'

  /* Get UNIXROOT and check whether it is valid */
  option.unixroot = CheckUnixroot(option.unixroot)
  if option.unixroot = '' then
    call Terminate 1, 'UNIXROOT invalid'
  else
    if option.verbose then say "UNIXROOT is '" || option.unixroot || "'"

  /* Read the tmp dir */
  option.tmpdir = GetTmpdir()
  if option.tmpdir = '' then
    call Terminate 1, 'TMPDIR invalid'
  else
    if option.verbose then say "TMPDIR is '" || option.tmpdir || "'"

  /* change slashes into backslashes */
  option.tmp = ChangeChar(option.tmpdir, '/', '\')


  if option.batch then
    call Terminate 1, 'batch mode not yet supported'

return /* End of ParseCommandLine */


DisplayHelp: PROCEDURE EXPOSE Pk CopyRight const.Usage
/* Prints help information. */
  say Pk
  say CopyRight
  say ''
  say const.Usage
  say ''
/*  say "Specific Program Options:"
  say ""*/

  say 'General program options:'
  say ' -b, --batch       Batch mode. The program succeeds or fails but never asks'
  say '                   the user. This option can not be combined with --wait.'
  say ' -d, --debug       Debug mode.'
  say ' -f, --force       The program tries to perform all actions and never asks'
  say '                   the user.'
  say ' -l, --list        Demo mode. Does not perform any changes.'
  say ' -q, --quiet       Quiet mode. This option can not be combined with --verbose.'
  say ' -r, --root ROOT   Allows to overwrite the UNIXROOT environment variable.'
  say ' -v, --verbose     Verbose mode. This option can not be combined with --quiet.'
  say ' -w, --wait        Wait mode. The program does not terminate before the user'
  say '                   has pressed any key. This option can not be combined with'
  say '                   --batch.'
  say '     --help        Prints help information.'
  say '     --version     Prints version information.'

  say ''
  say ''
exit /* End of DisplayHelp */

DisplayVersion: PROCEDURE EXPOSE Pk CopyRight Revision
/* Prints version info. */
  say Pk
  say CopyRight
  say Revision
  say ''
  say 'This is free software; see the source for copying conditions.'
  say 'There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A'
  say 'PARTICULAR PURPOSE.'
  say ''
exit /* End of DisplayVersion */


Terminate: PROCEDURE EXPOSE true false
/* Terminate with error message. */
  parse arg errorcode, message
  call lineout 'stderr:', message
  exit errorcode
return




/*
**********************************************************************
***** Exception handling *********************************************
**********************************************************************
*/

Exception.Syntax:
/* Syntax error */
  call lineout 'stderr:', 'Syntax error' || rc || '(' || ERRORTEXT(rc) || ') in line ' || SIGL || ':'
  call lineout 'stderr:', strip(SOURCELINE(SIGL))
exit /* End of Exception.Syntax */


Exception.NoValue:
/* Variable not initialized, rc might be undefined. */
  call lineout 'stderr:', 'Variable "' || CONDITION("D") || '" is not initialized in line ' || SIGL || ':'
  call lineout 'stderr:', strip(SOURCELINE(SIGL))
exit /* End of Exception.NoValue */


Exception.Error: PROCEDURE EXPOSE rc SIGL Catch.Error
/* Error while executing external command. */
  if symbol("Catch.Error") <> 'VAR' then Catch.Error = 0
  
  if Catch.Error < 2 then do /* Do not print any message if Catch.Error is 0 or 1. */
    call lineout 'stderr:', 'Error while executing external program in line ' || SIGL || ':'
    call lineout 'stderr:', strip(SOURCELINE(SIGL),,)
    call lineout 'stderr:', 'Error code is ' || rc
  end

  /* Do not terminate if this exception has to be caught. */
  if Catch.Error = 0 then exit
return /* End of Exception.Error */


Exception.Failure: PROCEDURE EXPOSE rc SIGL Catch.Failure
/* Unknown command, SIGL might be undefined. */
  if symbol("Catch.Failure") <> 'VAR' then Catch.Failure = 0

  if Catch.Failure < 2 then do /* Do not print any message if Catch.Failure is 0 or 1. */
    call lineout 'stderr:', 'System error ' || rc || ': Unknown command in line ' || SIGL || ':'
    call lineout 'stderr:', strip(SOURCELINE(SIGL),,)
  end

  /* Do not terminate if this exception has to be caught. */
  if Catch.Failure = 0 then exit
return /* Ende von "Fehler.Failure" */


Exception.Halt: PROCEDURE
/* Ctrl+C */
/* The signal handler is disabled so that another Ctrl+C does not lead to an endless
   loop where Ctrl+C is not accepted anymore. */
  CALL OFF HALT
  do forever
    call charout 'stdout:', 'You have pressed Crtl+C! Do you want to terminate the program (y/n)? '
    /* char Ch; */
    Ch = linein(,,)
    if Ch = 'J' | Ch = 'j' then call Terminate 3, 'Program aborted by user'
    else if Ch = 'N' | Ch = 'n' then do   /* Resume */
      CALL ON HALT NAME Exception.Halt    /* Activate signal handler */
      return
    end
  end /* do forever */

exit /* End of Exception.Halt */


Exception.NotReady: PROCEDURE EXPOSE rc SIGL Catch.NotReady
/* I/O error */
  if symbol("Catch.NotReady") <> 'VAR' then Catch.NotReady = 0
  
  if Catch.NotReady < 2 then do /* Do not print any message if Catch.NotReady is 2. */
    call lineout 'stderr:', 'I/O error for file/device "' || CONDITION("D") || '" in line ' || SIGL || ':'
    call lineout 'stderr:', strip(SOURCELINE(SIGL),,)
    if symbol("rc") = 'VAR' then 
      call lineout 'stderr:', 'Error code is ' || rc
  end
  
  /* Do not terminate if this exception has to be caught. */
  if Catch.NotReady = 0 then exit 1
return /* End of Exception.NotReady */


